# Copyright (c) 2022 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

"""Various types and definitions for the Haltium platform."""

from __future__ import annotations

import enum
from dataclasses import dataclass
from typing import Optional

from ._common import get_field, update_field


@enum.unique
class Product(enum.Enum):
    """Enumeration of supported product names in the Haltium family."""

    NRF54H20 = enum.auto()
    NRF9280 = enum.auto()

    @classmethod
    def _missing_(cls, value: object) -> Optional[Product]:
        """Custom enum behavior to support case insensitive values."""
        if not isinstance(value, str):
            return None
        return cls.__members__.get(value.upper())

    @property
    def product_code(self) -> ProductCode:
        match self:
            case Product.NRF54H20:
                return ProductCode(part_number=0x16, revision=0x2)
            case Product.NRF9280:
                return ProductCode(part_number=0x12, revision=0x1)
            case _:
                raise ValueError(f"No product code found for {self}")


@dataclass
class ProductCode:
    part_number: int
    revision: int


@enum.unique
class AddressOffset(enum.IntEnum):
    """Address bit offsets, defined by Address Format of the product specification."""

    REGION = 29
    SECURITY = 28
    DOMAINID = 24
    ADDR = 23


def secure_address_get(address: int) -> int:
    """Get the TrustZone secure address for the given address"""
    addr = Address(address)
    addr.security = True
    return int(addr)


@enum.unique
class AddressRegion(enum.IntEnum):
    """Address regions, defined by Address Format of the product specification."""

    PROGRAM = 0
    DATA = 1
    PERIPHERAL = 2
    EXT_XIP = 3
    EXT_XIP_ENCRYPTED = 4
    STM = 5
    CPU = 7

    @classmethod
    def from_address(cls, address: int) -> AddressRegion:
        """Get the address region of an address."""
        return Address(address).region


ADDRESS_REGION_POS = 29
ADDRESS_REGION_MASK = 0x7 << ADDRESS_REGION_POS
ADDRESS_SECURITY_POS = 28
ADDRESS_SECURITY_MASK = 0x1 << ADDRESS_SECURITY_POS
ADDRESS_DOMAIN_POS = 24
ADDRESS_DOMAIN_MASK = 0xF << ADDRESS_DOMAIN_POS
ADDRESS_BUS_POS = 16
ADDRESS_BUS_MASK = 0xFF << ADDRESS_BUS_POS
ADDRESS_SLAVE_POS = 12
ADDRESS_SLAVE_MASK = 0xF << ADDRESS_SLAVE_POS
ADDRESS_PERIPHID_POS = 12
ADDRESS_PERIPHID_MASK = 0x7FF << ADDRESS_PERIPHID_POS
ADDRESS_SPACE_POS = 0
ADDRESS_PROGRAM_DATA_SPACE_MASK = 0xFF_FFFF
ADDRESS_PERIPHERAL_SPACE_MASK = 0xFFF
ADDRESS_DEFAULT_SPACE_MASK = 0x1FFF_FFFF


# Regions that have domain ID and security fields
_HAS_DOMAIN_SECURITY = [
    AddressRegion.PROGRAM,
    AddressRegion.DATA,
    AddressRegion.PERIPHERAL,
    AddressRegion.STM,
]

# Regions that have the peripheral address format
_HAS_PERIPH_BITS = [
    AddressRegion.PERIPHERAL,
    AddressRegion.STM,
]


class Address:
    """Helper for working with addresses on the Haltium products."""

    def __init__(self, value: int = 0) -> None:
        self._val = value

    def __repr__(self) -> str:
        if self.region in _HAS_DOMAIN_SECURITY:
            domain_sec_str = (
                f", domain={self.domain.name} ({int(self.domain)}), "
                f"security={self.security}"
            )
        else:
            domain_sec_str = ""

        if self.region in _HAS_PERIPH_BITS:
            periph_bits_str = (
                f", bus={self.bus} (0b{self.bus:09_b}), "
                f"slave_index={self.slave_index} (0b{self.slave_index:09_b})"
            )
        else:
            periph_bits_str = ""

        field_str = (
            f"region={self.region.name} ({int(self.region)}){domain_sec_str}{periph_bits_str}, "
            f"address_space=0x{self.address_space:_x}"
        )

        return f"{type(self).__name__}({field_str})"

    def __str__(self) -> str:
        return repr(self)

    @property
    def region(self) -> AddressRegion:
        """Address region."""
        return AddressRegion(
            get_field(self._val, ADDRESS_REGION_POS, ADDRESS_REGION_MASK)
        )

    @region.setter
    def region(self, new: int) -> None:
        self._val = update_field(
            self._val, new, ADDRESS_REGION_POS, ADDRESS_REGION_MASK
        )

    @property
    def security(self) -> bool:
        """Address security (only present in some regions)."""
        self._check_has_security()
        return bool(get_field(self._val, ADDRESS_SECURITY_POS, ADDRESS_SECURITY_MASK))

    @security.setter
    def security(self, new: bool) -> None:
        self._check_has_security()
        self._val = update_field(
            self._val, int(new), ADDRESS_SECURITY_POS, ADDRESS_SECURITY_MASK
        )

    def _check_has_security(self) -> None:
        self._check_region_has_field(_HAS_DOMAIN_SECURITY, "security bit")

    @property
    def domain(self) -> DomainID:
        """Address domain ID (only present in some regions)."""
        self._check_has_domain_id()
        return DomainID(get_field(self._val, ADDRESS_DOMAIN_POS, ADDRESS_DOMAIN_MASK))

    @domain.setter
    def domain(self, new: DomainID | int) -> None:
        self._check_has_domain_id()
        self._val = update_field(
            self._val, new, ADDRESS_DOMAIN_POS, ADDRESS_DOMAIN_MASK
        )

    def _check_has_domain_id(self) -> None:
        self._check_region_has_field(_HAS_DOMAIN_SECURITY, "domain ID")

    @property
    def bus(self) -> int:
        """Bus ID (only present in some regions)."""
        self._check_has_bus()
        return get_field(self._val, ADDRESS_BUS_POS, ADDRESS_BUS_MASK)

    @bus.setter
    def bus(self, new: int) -> None:
        self._check_has_bus()
        self._val = update_field(self._val, new, ADDRESS_BUS_POS, ADDRESS_BUS_MASK)

    def _check_has_bus(self) -> None:
        self._check_region_has_field(_HAS_PERIPH_BITS, "Peripheral/APB bus number")

    @property
    def slave_index(self) -> int:
        """Slave index (only present in some regions)."""
        self._check_has_slave_index()
        return get_field(self._val, ADDRESS_SLAVE_POS, ADDRESS_SLAVE_MASK)

    @slave_index.setter
    def slave_index(self, new: int) -> None:
        self._check_has_slave_index()
        self._val = update_field(self._val, new, ADDRESS_SLAVE_POS, ADDRESS_SLAVE_MASK)

    def _check_has_slave_index(self) -> None:
        self._check_region_has_field(_HAS_PERIPH_BITS, "Peripheral/APB slave index")

    @property
    def address_space(self) -> int:
        """Internal address space address (semantics depend on the region)."""
        match self.region:
            case AddressRegion.PROGRAM | AddressRegion.DATA:
                return get_field(
                    self._val, ADDRESS_SPACE_POS, ADDRESS_PROGRAM_DATA_SPACE_MASK
                )
            case AddressRegion.PERIPHERAL | AddressRegion.STM:
                return get_field(
                    self._val, ADDRESS_SPACE_POS, ADDRESS_PERIPHERAL_SPACE_MASK
                )
            case _:
                return get_field(
                    self._val, ADDRESS_SPACE_POS, ADDRESS_DEFAULT_SPACE_MASK
                )

    @address_space.setter
    def address_space(self, new: int) -> None:
        match self.region:
            case AddressRegion.PROGRAM | AddressRegion.DATA:
                self._val = update_field(
                    self._val, new, ADDRESS_SPACE_POS, ADDRESS_PROGRAM_DATA_SPACE_MASK
                )
            case AddressRegion.PERIPHERAL | AddressRegion.STM:
                self._val = update_field(
                    self._val, new, ADDRESS_SPACE_POS, ADDRESS_PERIPHERAL_SPACE_MASK
                )
            case _:
                self._val = update_field(
                    self._val, new, ADDRESS_SPACE_POS, ADDRESS_DEFAULT_SPACE_MASK
                )

    def _check_region_has_field(
        self, valid_regions: list[AddressRegion], field_name: str
    ) -> None:
        if self.region not in valid_regions:
            raise ValueError(
                f"{field_name} is not defined for address region {self.region.name}"
            )

    def __int__(self) -> int:
        return self._val


def peripheral_id_get(periph_address: int) -> int:
    """Get the peripheral ID of a peripheral address."""
    return get_field(periph_address, ADDRESS_PERIPHID_POS, ADDRESS_PERIPHID_MASK)


@enum.unique
class DomainID(enum.IntEnum):
    """Domain IDs in Haltium products."""

    RESERVED = 0
    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELLCORE = 4
    CELLDSP = 5
    CELLRF = 6
    ISIMCORE = 7
    GLOBALFAST = 12
    GLOBALSLOW = 13
    GLOBAL_ = 14
    GLOBAL = 15

    @classmethod
    def from_address(cls, address: int) -> DomainID:
        """Get the domain ID of an address."""
        return cls((address >> AddressOffset.DOMAINID) & 0xF)

    @classmethod
    def from_processor(cls, processor: ProcessorID | int) -> DomainID:
        """Get the domain ID corresponding to a processor ID."""
        processor_domain = {
            ProcessorID.SECURE: cls.SECURE,
            ProcessorID.APPLICATION: cls.APPLICATION,
            ProcessorID.RADIOCORE: cls.RADIOCORE,
            ProcessorID.CELLCORE: cls.CELLCORE,
            ProcessorID.CELLDSP: cls.CELLDSP,
            ProcessorID.CELLRF: cls.CELLRF,
            ProcessorID.ISIMCORE: cls.ISIMCORE,
            ProcessorID.SYSCTRL: cls.GLOBALFAST,
            ProcessorID.PPR: cls.GLOBALSLOW,
            ProcessorID.FLPR: cls.GLOBAL_,
        }
        return processor_domain[ProcessorID(processor)]

    @property
    def c_enum(self) -> str:
        return f"NRF_DOMAIN_{self.name.upper()}"


@enum.unique
class OwnerID(enum.IntEnum):
    """Enumeration of ownership IDs in haltium products."""

    NONE = 0
    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELL = 4
    ISIMCORE = 5
    SYSCTRL = 8

    @classmethod
    def from_domain(cls, domain: DomainID | int) -> OwnerID:
        """Get the owner ID corresponding to a domain ID."""
        domain_owner = {
            DomainID.SECURE: cls.SECURE,
            DomainID.APPLICATION: cls.APPLICATION,
            DomainID.RADIOCORE: cls.RADIOCORE,
            DomainID.CELLCORE: cls.CELL,
            DomainID.CELLDSP: cls.CELL,
            DomainID.CELLRF: cls.CELL,
            DomainID.ISIMCORE: cls.ISIMCORE,
            DomainID.GLOBALFAST: cls.SYSCTRL,
        }
        return domain_owner[DomainID(domain)]

    @classmethod
    def from_processor(cls, processor: ProcessorID | int) -> OwnerID:
        """Get the owner ID corresponding to a processor ID."""
        return cls.from_domain(DomainID.from_processor(processor))

    @property
    def c_enum(self) -> str:
        return f"NRF_OWNER_{self.name.upper()}"


@enum.unique
class ProcessorID(enum.IntEnum):
    """Processor IDs in haltium products."""

    SECURE = 1
    APPLICATION = 2
    RADIOCORE = 3
    CELLCORE = 4
    CELLDSP = 5
    CELLRF = 6
    ISIMCORE = 7
    BBPR = 11
    SYSCTRL = 12
    PPR = 13
    FLPR = 14

    @classmethod
    def from_domain(cls, domain: DomainID | int) -> ProcessorID:
        """Get the processor ID corresponding to a domain ID."""
        domain_processor = {
            DomainID.SECURE: cls.SECURE,
            DomainID.APPLICATION: cls.APPLICATION,
            DomainID.RADIOCORE: cls.RADIOCORE,
            DomainID.CELLCORE: cls.CELLCORE,
            DomainID.CELLDSP: cls.CELLDSP,
            DomainID.CELLRF: cls.CELLRF,
            DomainID.ISIMCORE: cls.ISIMCORE,
            DomainID.GLOBALFAST: cls.SYSCTRL,
            DomainID.GLOBALSLOW: cls.PPR,
            DomainID.GLOBAL_: cls.FLPR,
        }
        return domain_processor[DomainID(domain)]

    @property
    def c_enum(self) -> str:
        return f"NRF_PROCESSOR_{self.name.upper()}"


@enum.unique
class Ctrlsel(enum.IntEnum):
    """
    Enumeration of GPIO.PIN_CNF[n].CTRLSEL values.
    The list here may not be exhaustive.
    """

    GPIO = 0
    VPR_GRC = 1
    CAN_PWM_I3C = 2
    SERIAL0 = 3
    EXMIF_RADIO_SERIAL1 = 4
    CAN_TDM_SERIAL2 = 5
    CAN = 6
    TND = 7


# Default CTRLSEL value indicating that CTRLSEL should not be used
CTRLSEL_DEFAULT = Ctrlsel.GPIO



# Copyright (c) 2025 Nordic Semiconductor ASA
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import importlib.resources
import pickle
from collections import defaultdict
from enum import auto, Flag
from functools import lru_cache
from itertools import chain
from pathlib import Path
from typing import Iterable, Optional, Sequence, TYPE_CHECKING

import svd
from intelhex import IntelHex

if TYPE_CHECKING:
    from devicetree import edtlib

from . import log, uicr, __version__
from .platform import (
    Address,
    AddressRegion,
    CTRLSEL_DEFAULT,
    DomainID,
    OwnerID,
    peripheral_id_get,
    ProcessorID,
)
from ._common import get_field
from ._svd_util import SvdEnum


class MigrateFlag(Flag):
    DEFAULTS = auto()

    @classmethod
    def all(cls) -> MigrateFlag:
        return ~MigrateFlag(0)

    @classmethod
    def default(cls) -> MigrateFlag:
        return cls.all()


def uicr_migrate_hex_files_to_periphconf(
    uicr_hex_files: list[Path],
    edt_pickle_file: Optional[Path] = None,
    flags: MigrateFlag = MigrateFlag.default(),
) -> str:
    if edt_pickle_file is not None:
        with edt_pickle_file.open("rb") as file:
            dt = pickle.load(file)

        periph_annotation_lookup = _build_periph_annotation_dt(dt)
    else:
        periph_annotation_lookup = {}

    to_migrate = []

    for uicr_hex_file in uicr_hex_files:
        ihex = IntelHex()
        with uicr_hex_file.open("r", encoding="utf-8") as fp:
            ihex.loadhex(fp)

        uicr_bytes = bytearray(ihex.tobinstr())
        to_migrate.append(uicr.Uicr.from_bytes(uicr_bytes))

    return uicr_migrate(
        to_migrate,
        flags=flags,
        periph_annotation_lookup=periph_annotation_lookup,
    )


def _build_periph_annotation_dt(dt: edtlib.EDT) -> dict[int, str]:
    lut = {}

    for node in dt.nodes:
        if not node.regs:
            continue

        addr = node.regs[0].addr
        if node.labels:
            lut[addr] = node.labels[0]
        else:
            lut[addr] = node.path

    return lut


def uicr_migrate(
    to_migrate: Sequence[uicr.Uicr],
    flags: MigrateFlag = MigrateFlag.default(),
    periph_annotation_lookup: dict[int, str] = {},
) -> str:
    header_lines = [
        f"/* Generated by nrf-regtool version {__version__} */",
        "#include <uicr/uicr.h>",
        "",
    ]

    ipcmap_idx = 0

    # For grouping lines by type so that e.g. GPIO settings from two different UICRs are grouped
    # together (the group name is arbitrary).
    pconf_lines_by_type = defaultdict(list)

    for builder in to_migrate:
        uicr_svd = builder._uicr

        default_domain = DomainID(uicr_svd["HEADER"]["DOMAIN"]["DOMAIN"].content)
        default_processor = ProcessorID.from_domain(default_domain)
        default_owner = OwnerID.from_processor(default_processor)

        for reg in uicr_svd["IPCMAP"]:
            if not reg.modified:
                break

            source_domain = DomainID(reg["DOMAINIDSOURCE"].content)
            source_ch = reg["IPCTCHSOURCE"].content
            sink_domain = DomainID(reg["DOMAINIDSINK"].content)
            sink_ch = reg["IPCTCHSINK"].content

            pconf_lines_by_type["IPCMAP"].append(
                f"/* {source_domain.name} IPCT ch. {source_ch} => "
                f"{sink_domain.name} IPCT ch. {sink_ch} */"
            )
            pconf_lines_by_type["IPCMAP"].append(
                (
                    "UICR_IPCMAP_CHANNEL_CFG("
                    f"{ipcmap_idx}, "
                    f"{source_domain.c_enum}, {source_ch}, "
                    f"{sink_domain.c_enum}, {sink_ch});"
                )
            )
            ipcmap_idx += 1

        for reg in uicr_svd["GPIOTE"]:
            if not reg["INSTANCE"].modified:
                break

            address = reg["INSTANCE"].content
            instance_name = periph_annotation_lookup.get(address, _c_hex_addr(address))
            spu_address = _get_spu_addr_for_periph(address)
            spu_name = _SPU_ADDR_TO_NAME[spu_address]

            for i, secure in _iter_owned_secure(reg["CH"]["OWN"], reg["CH"]["SECURE"]):
                if MigrateFlag.DEFAULTS not in flags and (secure, default_owner) == (
                    True,
                    OwnerID.APPLICATION,
                ):
                    continue

                pconf_lines_by_type["SPU_GPIOTE_CH"].append(
                    f"/* {spu_name} feature configuration for {instance_name} ch. {i} */"
                )
                pconf_lines_by_type["SPU_GPIOTE_CH"].append(
                    (
                        "UICR_SPU_FEATURE_GPIOTE_CH_SET("
                        f"{_c_hex_addr(spu_address)}, "
                        f"0, {i}, {_c_bool(secure)}, {default_owner.c_enum});"
                    )
                )

        for reg in uicr_svd["IPCT"]:
            if not reg["INSTANCE"].modified:
                break

            address = reg["INSTANCE"].content
            instance_name = periph_annotation_lookup.get(address, _c_hex_addr(address))
            spu_address = _get_spu_addr_for_periph(address)
            spu_name = _SPU_ADDR_TO_NAME[spu_address]

            for i, secure in _iter_owned_secure(reg["CH"]["OWN"], reg["CH"]["SECURE"]):
                if MigrateFlag.DEFAULTS not in flags and (secure, default_owner) == (
                    True,
                    OwnerID.APPLICATION,
                ):
                    continue

                pconf_lines_by_type["SPU_IPCT_CH"].append(
                    f"/* {spu_name} feature configuration for {instance_name} ch. {i} */"
                )
                pconf_lines_by_type["SPU_IPCT_CH"].append(
                    (
                        "UICR_SPU_FEATURE_IPCT_CH_SET("
                        f"{_c_hex_addr(spu_address)}, "
                        f"{i}, {_c_bool(secure)}, {default_owner.c_enum});"
                    )
                )

            for i, secure in _iter_owned_secure(
                reg["INTERRUPT"]["OWN"], reg["INTERRUPT"]["SECURE"]
            ):
                if MigrateFlag.DEFAULTS not in flags and (secure, default_owner) == (
                    True,
                    OwnerID.APPLICATION,
                ):
                    continue

                pconf_lines_by_type["SPU_IPCT_INT"].append(
                    f"/* {spu_name} feature configuration for {instance_name} interrupt {i} */"
                )
                pconf_lines_by_type["SPU_IPCT_INT"].append(
                    (
                        "UICR_SPU_FEATURE_IPCT_INTERRUPT_SET("
                        f"{_c_hex_addr(spu_address)}, "
                        f"{i}, {_c_bool(secure)}, {default_owner.c_enum});"
                    )
                )

        for reg in uicr_svd["DPPI"]:
            if not reg["INSTANCE"].modified:
                break

            address = reg["INSTANCE"].content
            instance_name = _DPPIC_ADDR_TO_NAME[address]
            if instance_name == "DPPIC130":
                continue

            spu_address = _get_spu_addr_for_periph(address)
            spu_name = _SPU_ADDR_TO_NAME[spu_address]

            for i, secure in _iter_owned_secure(reg["CH"]["OWN"], reg["CH"]["SECURE"]):
                if MigrateFlag.DEFAULTS not in flags and (secure, default_owner) == (
                    True,
                    OwnerID.APPLICATION,
                ):
                    continue

                pconf_lines_by_type["SPU_DPPI_CH"].append(
                    f"/* {spu_name} feature configuration for {instance_name} ch. {i} */"
                )
                pconf_lines_by_type["SPU_DPPI_CH"].append(
                    (
                        "UICR_SPU_FEATURE_DPPIC_CH_SET("
                        f"{_c_hex_addr(spu_address)}, "
                        f"{i}, {_c_bool(secure)}, {default_owner.c_enum});"
                    )
                )

            for i, secure in _iter_owned_secure(
                reg["CHG"]["OWN"], reg["CHG"]["SECURE"]
            ):
                if MigrateFlag.DEFAULTS not in flags and (secure, default_owner) == (
                    True,
                    OwnerID.APPLICATION,
                ):
                    continue

                pconf_lines_by_type["SPU_DDPI_CHG"].append(
                    f"/* {spu_name} feature configuration for {instance_name} ch. group {i} */"
                )
                pconf_lines_by_type["SPU_DPPI_CHG"].append(
                    (
                        "UICR_SPU_FEATURE_DPPIC_CHG_SET("
                        f"{_c_hex_addr(spu_address)}, "
                        f"{i}, {_c_bool(secure)}, {default_owner.c_enum});"
                    )
                )

            reg_link = reg["CH"]["LINK"]
            for ch, (enabled, direction) in enumerate(
                zip(reg_link["EN"].values(), reg_link["DIR"].values())
            ):
                if enabled.content_enum != "Enabled":
                    continue

                local_ppib_name, local_ppib_ch_map = _DPPIC_TO_PPIB[instance_name]
                local_ppib_ch = local_ppib_ch_map[ch]

                remote_ppib_name, remote_ppib_ch_map = _PPIB_TO_PPIB[local_ppib_name]
                remote_ppib_ch = remote_ppib_ch_map[local_ppib_ch]

                if direction.content_enum == "Source":
                    sub_name, sub_ch = local_ppib_name, local_ppib_ch
                    pub_name, pub_ch = remote_ppib_name, remote_ppib_ch
                else:
                    sub_name, sub_ch = remote_ppib_name, remote_ppib_ch
                    pub_name, pub_ch = local_ppib_name, local_ppib_ch

                sub_addr = _PPIB_NAME_TO_ADDR[sub_name]
                sub_final_name = periph_annotation_lookup.get(sub_addr, sub_name)
                pub_addr = _PPIB_NAME_TO_ADDR[pub_name]
                pub_final_name = periph_annotation_lookup.get(pub_addr, pub_name)

                pconf_lines_by_type["DPPI"].append(
                    f"/* {sub_final_name} ch. {sub_ch} => {pub_final_name} ch. {pub_ch} */"
                )
                pconf_lines_by_type["DPPI"].append(
                    f"UICR_PPIB_SUBSCRIBE_SEND_ENABLE({_c_hex_addr(sub_addr)}, {sub_ch});"
                )
                pconf_lines_by_type["DPPI"].append(
                    f"UICR_PPIB_PUBLISH_RECEIVE_ENABLE({_c_hex_addr(pub_addr)}, {pub_ch});"
                )

        grtc_spu_address = _get_spu_addr_for_periph(_GRTC_ADDR)
        grtc_spu_name = _SPU_ADDR_TO_NAME[grtc_spu_address]

        for i, secure in _iter_owned_secure(
            uicr_svd["GRTC"]["CC"]["OWN"], uicr_svd["GRTC"]["CC"]["SECURE"]
        ):
            if MigrateFlag.DEFAULTS in flags or (
                not secure or default_owner != OwnerID.APPLICATION
            ):
                pconf_lines_by_type["SPU_GRTC_CC"].append(
                    f"/* {grtc_spu_name} feature configuration for GRTC CC{i} */"
                )
                pconf_lines_by_type["SPU_GRTC_CC"].append(
                    (
                        "UICR_SPU_FEATURE_GRTC_CC_SET("
                        f"{_c_hex_addr(grtc_spu_address)}, "
                        f"{i}, {_c_bool(secure)}, {default_owner.c_enum});"
                    )
                )

        for i, reg in enumerate(uicr_svd["GPIO"]):
            if not reg["INSTANCE"].modified:
                break

            address = reg["INSTANCE"].content
            instance_name = periph_annotation_lookup.get(address, _c_hex_addr(address))
            spu_address = _get_spu_addr_for_periph(address)
            spu_name = _SPU_ADDR_TO_NAME[spu_address]
            gpio_port = _get_gpio_port_num(address)

            for pin, own in enumerate(reg["OWN"].values()):
                if own.content_enum != "Own":
                    continue

                secure = reg["SECURE"][f"PIN_{i}"].content_enum == "Secure"

                if MigrateFlag.DEFAULTS in flags or (
                    not secure or default_owner != OwnerID.APPLICATION
                ):
                    pconf_lines_by_type["SPU_GPIO"].append(
                        f"/* {spu_name} feature configuration for {instance_name}, "
                        f"P{gpio_port}.{pin} */"
                    )
                    pconf_lines_by_type["SPU_GPIO"].append(
                        (
                            "UICR_SPU_FEATURE_GPIO_PIN_SET("
                            f"{_c_hex_addr(spu_address)}, "
                            f"{gpio_port}, {pin}, {_c_bool(secure)}, {default_owner.c_enum});"
                        )
                    )

                ctrlsel = uicr_svd["GPIO_PIN"][i]["CTRLSEL"][pin]
                if not ctrlsel.modified:
                    continue

                ctrlsel_val = ctrlsel.content

                if MigrateFlag.DEFAULTS in flags or ctrlsel_val != CTRLSEL_DEFAULT:
                    pconf_lines_by_type["GPIO"].append(
                        f"/* {instance_name} - P{gpio_port}.{pin} CTRLSEL = {ctrlsel_val} */"
                    )
                    pconf_lines_by_type["GPIO"].append(
                        f"UICR_GPIO_PIN_CNF_CTRLSEL_SET({_c_hex_addr(address)}, "
                        f"{pin}, {ctrlsel.content});"
                    )

        for reg in uicr_svd["PERIPH"]:
            reg_cfg = reg["CONFIG"]
            if not reg_cfg.modified:
                break

            address = reg_cfg["ADDRESS"].content << reg_cfg["ADDRESS"].bit_offset
            irq_processor = ProcessorID(reg_cfg["PROCESSOR"].content)
            secure = reg_cfg["SECURE"].content_enum == "Secure"
            dma_secure = reg_cfg["DMASEC"].content_enum == "Secure"

            pconf_lines_by_type["SPU_PERIPH"].extend(
                _make_spu_periph_perm_set(
                    periph_addr=address,
                    secure=secure,
                    dma_secure=dma_secure,
                    owner=default_owner,
                    flags=flags,
                    periph_annotation_lookup=periph_annotation_lookup,
                )
            )
            pconf_lines_by_type["IRQMAP"].extend(
                _make_irqmap_irq_sink_set(
                    periph_addr=address,
                    irq_processor=irq_processor,
                    flags=flags,
                    periph_annotation_lookup=periph_annotation_lookup,
                )
            )

    full_lines = header_lines + list(chain.from_iterable(pconf_lines_by_type.values()))

    return "\n".join(full_lines)


def _iter_owned_secure(
    reg_own: svd.Register, reg_sec: svd.Register
) -> Iterable[tuple[int, bool]]:
    for i, (own, sec) in enumerate(zip(reg_own.values(), reg_sec.values())):
        owned = not own.content
        if not owned:
            continue

        yield i, sec.content_enum == SvdEnum.SECURE


_SPU_PERIPH_PERM_OFFSET = 0x500
_SPU_PERIPH_PERM_DMA_NODMA = 0


def _make_spu_periph_perm_set(
    periph_addr: int,
    secure: bool,
    dma_secure: bool,
    owner: OwnerID,
    flags: MigrateFlag,
    periph_annotation_lookup: dict[int, str],
) -> list[str]:
    periph_name = periph_annotation_lookup.get(periph_addr, _c_hex_addr(periph_addr))
    pconf_defaults = _load_pconf_defaults()

    if periph_addr in (_VPR121_ADDR, _VPR130_ADDR):
        log.debug(f"skip SPU.PERIPH[n].PERM for VPR peripheral {periph_name}")
        return []

    parsed_addr = Address(periph_addr)
    if parsed_addr.region == AddressRegion.STM:
        log.debug(f"skip SPU.PERIPH[n].PERM for TDD peripheral {periph_name}")
        return []

    spu_address = _get_spu_addr_for_periph(periph_addr)
    spu_name = _SPU_ADDR_TO_NAME[spu_address]
    periph_slave_index = parsed_addr.slave_index

    default_value = pconf_defaults[
        spu_address + _SPU_PERIPH_PERM_OFFSET + 4 * periph_slave_index
    ]
    default_fields = {
        "SECUREMAPPING": get_field(default_value, 0, 0b11),
        "DMA": get_field(default_value, 2, 0b11 << 2),
        "SECATTR": bool(get_field(default_value, 4, 0b1 << 4)),
        "DMASEC": bool(get_field(default_value, 5, 0b1 << 5)),
        "OWNERID": get_field(default_value, 16, 0b1111 << 16),
        "OWNERPROG": get_field(default_value, 30, 0b1 << 30),
    }

    if default_fields["OWNERID"] in (OwnerID.SECURE, OwnerID.SYSCTRL):
        log.debug(
            f"skip unassignable SPU.PERIPH[{periph_slave_index}].PERM "
            f"(default owner is {default_fields['OWNERID']})"
        )
        return []

    # If there is no DMA, use the default value to avoid readback errors
    if default_fields["DMA"] == _SPU_PERIPH_PERM_DMA_NODMA:
        dma_secure = default_fields["DMASEC"]
        log.debug(
            f"use default DMASEC={dma_secure} for SPU.PERIPH[{periph_slave_index}].PERM (noDMA)"
        )

    if (secure, dma_secure, owner) == (
        default_fields["SECATTR"],
        default_fields["DMASEC"],
        default_fields["OWNERID"],
    ) and MigrateFlag.DEFAULTS not in flags:
        log.debug(
            f"skip default configuration of SPU.PERIPH[{periph_slave_index}].PERM "
            f"(default values are {default_fields['SECATTR']=}, {default_fields['DMASEC']=}, "
            f"{default_fields['OWNERID']=})"
        )
        return []

    return [
        f"/* {spu_name} configuration for {periph_name} */",
        (
            "UICR_SPU_PERIPH_PERM_SET("
            f"{_c_hex_addr(spu_address)}, {periph_slave_index}, "
            f"{_c_bool(secure)}, {_c_bool(dma_secure)}, {owner.c_enum});"
        ),
    ]


_IRQMAP_ADDR = 0x5F92_4000
_IRQMAP_IRQ_SINK_OFFSET = 0x400


def _make_irqmap_irq_sink_set(
    periph_addr: int,
    irq_processor: ProcessorID,
    flags: MigrateFlag,
    periph_annotation_lookup: dict[int, str],
) -> list[str]:
    periph_name = periph_annotation_lookup.get(periph_addr, _c_hex_addr(periph_addr))
    pconf_defaults = _load_pconf_defaults()

    if periph_addr == _TBM_ADDR:
        periph_irqn = _TBM_IRQN
    else:
        periph_irqn = peripheral_id_get(periph_addr)

    default_value = pconf_defaults[
        _IRQMAP_ADDR + _IRQMAP_IRQ_SINK_OFFSET + 4 * periph_irqn
    ]
    default_processor_int = get_field(default_value, 8, 0b1111 << 8)

    if default_processor_int in (0, ProcessorID.SECURE, ProcessorID.SYSCTRL):
        log.debug(
            f"skip unassignable IRQMAP.IRQ[{periph_irqn}].SINK config "
            f"(default processor is {default_processor_int})"
        )
        return []

    if irq_processor == default_processor_int and MigrateFlag.DEFAULTS not in flags:
        log.debug(
            f"skip default IRQMAP.IRQ[{periph_irqn}].SINK config "
            f"(default processor is {default_processor_int})"
        )
        return []

    return [
        f"/* {periph_name} IRQ => {irq_processor.name} */",
        f"UICR_IRQMAP_IRQ_SINK_SET({periph_irqn}, " f"{irq_processor.c_enum});",
    ]


@lru_cache(1)
def _load_pconf_defaults() -> dict[int, int]:
    pconf_hex_path = importlib.resources.files("nrfregtool.resources").joinpath(
        "nrf54h20_periphconf_defaults.hex"
    )

    with importlib.resources.as_file(pconf_hex_path) as hex_path:
        with open(hex_path, "r", encoding="utf-8") as fp:
            pconf_hex = IntelHex()
            pconf_hex.loadhex(fp)

    defaults = {}
    pconf_addresses = pconf_hex.addresses()
    for i in range(0, len(pconf_addresses), 4):
        addresses = [pconf_addresses[i + j] for j in range(4)]
        value_parts = [pconf_hex[a] for a in addresses]
        value = int.from_bytes(value_parts, "little")
        defaults[addresses[0]] = value

    return defaults


def _get_spu_addr_for_periph(periph_addr: int) -> int:
    address = Address(periph_addr)
    address.slave_index = 0
    address.address_space = 0
    # Usually the above is sufficient to get the SPU address, but some buses
    # are special due to having > 16 slaves and need to be handled here.
    address.bus = _SPU_ADDRESS_BUS_REMAPPING.get(address.bus, address.bus)

    return int(address)


_SPU_ADDRESS_BUS_REMAPPING = {
    # Both of these bus IDs represent APB32 and should use the same SPU instance with bus ID 146.
    147: 146,
}


def _get_gpio_port_num(gpio_addr: int) -> int:
    return (gpio_addr - _P0_ADDR) // _GPIO_SIZE


_P0_ADDR = 0x5F93_8000
_GPIO_SIZE = 0x200

_GRTC_ADDR = 0x5F99_C000

_TDDCONF_ADDR = 0xBF00_1000

_TBM_ADDR = 0xBF00_3000
_TBM_IRQN = 127

_VPR121_ADDR = 0x5F8D_4000
_VPR130_ADDR = 0x5F90_8000


def _c_hex_addr(address: int) -> str:
    return f"0x{address:08x}UL"


def _c_bool(value: bool) -> str:
    return "true" if value else "false"


_SPU_ADDR_TO_NAME = {
    0x5F08_0000: "SPU110",
    0x5F09_0000: "SPU111",
    0x5F8C_0000: "SPU120",
    0x5F8D_0000: "SPU121",
    0x5F8E_0000: "SPU122",
    0x5F90_0000: "SPU130",
    0x5F92_0000: "SPU131",
    0x5F98_0000: "SPU132",
    0x5F99_0000: "SPU133",
    0x5F9A_0000: "SPU134",
    0x5F9B_0000: "SPU135",
    0x5F9C_0000: "SPU136",
    0x5F9D_0000: "SPU137",
}

_DPPIC_ADDR_TO_NAME = {
    0x5F8E_1000: "DPPIC120",
    0x5F92_2000: "DPPIC130",
    0x5F98_1000: "DPPIC131",
    0x5F99_1000: "DPPIC132",
    0x5F9A_1000: "DPPIC133",
    0x5F9B_1000: "DPPIC134",
    0x5F9C_1000: "DPPIC135",
    0x5F9D_1000: "DPPIC136",
}

_PPIB_NAME_TO_ADDR = {
    "PPIB121": 0x5F8E_F000,
    "PPIB130": 0x5F92_5000,
    "PPIB131": 0x5F92_6000,
    "PPIB132": 0x5F98_D000,
    "PPIB133": 0x5F99_D000,
    "PPIB134": 0x5F9A_D000,
    "PPIB135": 0x5F9B_D000,
    "PPIB136": 0x5F9C_D000,
    "PPIB137": 0x5F9D_D000,
}

_DPPIC_TO_PPIB = {
    "DPPIC120": ("PPIB121", range(0, 8)),
    "DPPIC131": ("PPIB132", range(0, 8)),
    "DPPIC132": ("PPIB133", range(0, 8)),
    "DPPIC133": ("PPIB134", range(0, 8)),
    "DPPIC134": ("PPIB135", range(0, 8)),
    "DPPIC135": ("PPIB136", range(0, 8)),
    "DPPIC136": ("PPIB137", range(0, 8)),
}

_PPIB_TO_PPIB = {
    "PPIB132": ("PPIB130", range(0, 8)),
    "PPIB133": ("PPIB130", range(8, 16)),
    "PPIB134": ("PPIB130", range(16, 24)),
    "PPIB135": ("PPIB130", range(24, 32)),
    "PPIB136": ("PPIB131", range(0, 8)),
    "PPIB137": ("PPIB131", range(8, 16)),
    "PPIB121": ("PPIB131", range(16, 24)),
}
